"""Client API for Nexus."""
import typing

import httpx

from qnexus.client.utils import read_token, write_token
from qnexus.config import get_config
from qnexus.exceptions import AuthenticationError

config = get_config()
_nexus_client: httpx.Client | None = None


class AuthHandler(httpx.Auth):
    """Custom nexus auth handler"""

    cookies: httpx.Cookies

    def __init__(self) -> None:
        self.cookies = httpx.Cookies()
        try:
            token = read_token(
                "refresh_token",
            )
            self.cookies.set("myqos_oat", token, domain=get_config().domain)
        except FileNotFoundError:
            pass  # Okay to ignore this as the user may log in later

        super().__init__()

    def auth_flow(
        self, request: httpx.Request
    ) -> typing.Generator[httpx.Request, httpx.Response, None]:
        self.cookies.set_cookie_header(request)
        response = yield request
        if response.status_code == 401:
            if self.cookies.get("myqos_oat") is None:
                try:
                    token = read_token(
                        "refresh_token",
                    )
                    self.cookies.set("myqos_oat", token, domain=get_config().domain)
                except FileNotFoundError as exc:
                    raise AuthenticationError(
                        "Not authenticated. Please run `qnx login` in your terminal."
                    ) from exc

            auth_response = yield self.build_refresh_request()
            if auth_response.status_code == 401:
                raise AuthenticationError(
                    "Not authenticated. Please run `qnx login` in your terminal."
                )

            auth_response.raise_for_status()
            self.cookies.extract_cookies(auth_response)

            write_token(
                "access_token",
                self.cookies.get("myqos_id", domain=get_config().domain) or "",
            )
            if request.headers.get("cookie"):
                request.headers.pop("cookie")
            self.cookies.set_cookie_header(request)
            yield request

    def build_refresh_request(self) -> httpx.Request:
        """Build the request for refreshing the id token."""
        self.cookies.delete("myqos_id")  # We need to delete the existing token first
        return httpx.Request(
            method="POST",
            url=f"{get_config().url}/auth/tokens/refresh",
            cookies=self.cookies,
        )


def get_nexus_client() -> httpx.Client:
    """Getter function for the nexus client."""
    global _nexus_client  # pylint: disable=global-statement
    if _nexus_client is None:
        _nexus_client = httpx.Client(
            base_url=config.url,
            auth=AuthHandler(),
            timeout=None,
            verify=config.httpx_verify,
        )
    return _nexus_client


def reload_client() -> None:
    """Reload the nexus client with new tokens."""
    global _nexus_client  # pylint: disable=global-statement
    _nexus_client = get_nexus_client()
    _nexus_client.cookies.clear()
    _nexus_client.auth.cookies.clear()  # type:ignore
    _nexus_client = httpx.Client(
        base_url=get_config().url,
        auth=AuthHandler(),
        timeout=None,
        verify=get_config().httpx_verify,
    )



"""Client API for results in Nexus."""

# https://staging.myqos.com/api-docs#/results


# def get():
#     pass


# def get_only():
#     pass


# def _fetch_by_id():
#     pass


# def update():
#     pass



"""Client API for HUGR in Nexus.

N.B. Nexus support for HUGR is experimental, and any HUGRs programs
uploaded to Nexus before stability is achieved might not work in the future.
"""

# pylint: disable=redefined-builtin
import base64
from datetime import datetime
from typing import Any, Union, cast
from uuid import UUID

from hugr.envelope import EnvelopeConfig, EnvelopeFormat
from hugr.package import Package

import qnexus.exceptions as qnx_exc
from qnexus.client import get_nexus_client
from qnexus.client.nexus_iterator import NexusIterator
from qnexus.client.utils import handle_fetch_errors
from qnexus.context import (
    get_active_project,
    merge_project_from_context,
    merge_properties_from_context,
)
from qnexus.models.annotations import Annotations, CreateAnnotations, PropertiesDict
from qnexus.models.filters import (
    CreatorFilter,
    FuzzyNameFilter,
    PaginationFilter,
    ProjectRefFilter,
    PropertiesFilter,
    ScopeFilter,
    ScopeFilterEnum,
    SortFilter,
    SortFilterEnum,
    TimeFilter,
)
from qnexus.models.references import DataframableList, HUGRRef, ProjectRef


class Params(
    SortFilter,
    PaginationFilter,
    FuzzyNameFilter,
    CreatorFilter,
    ProjectRefFilter,
    PropertiesFilter,
    TimeFilter,
    ScopeFilter,
):
    """Params for filtering HUGRs."""


# We can change the format and zstd when HUGR supports more options Since the
# header in the envelope encodes the config, Package.from_bytes will work
# without changes. We expect HUGR team to make other formats available during
# March 2025.
ENVELOPE_CONFIG = EnvelopeConfig(
    # As of hugr v0.11.3, the only format available is JSON
    format=EnvelopeFormat.JSON,
    # disable zstd compression for now
    zstd=None,
)


@merge_project_from_context
def get_all(  # pylint: disable=too-many-positional-arguments
    name_like: str | None = None,
    creator_email: list[str] | None = None,
    project: ProjectRef | None = None,
    properties: PropertiesDict | None = None,
    created_before: datetime | None = None,
    created_after: datetime | None = datetime(day=1, month=1, year=2023),
    modified_before: datetime | None = None,
    modified_after: datetime | None = None,
    sort_filters: list[SortFilterEnum] | None = None,
    page_number: int | None = None,
    page_size: int | None = None,
    scope: ScopeFilterEnum | None = None,
) -> NexusIterator[HUGRRef]:
    """Get a NexusIterator over HUGRs with optional filters."""

    params = Params(
        name_like=name_like,
        creator_email=creator_email,
        properties=properties,
        project=project,
        created_before=created_before,
        created_after=created_after,
        modified_before=modified_before,
        modified_after=modified_after,
        sort=SortFilter.convert_sort_filters(sort_filters),
        page_number=page_number,
        page_size=page_size,
        scope=scope,
    ).model_dump(by_alias=True, exclude_unset=True, exclude_none=True)

    return NexusIterator(
        resource_type="HUGR",
        nexus_url="/api/hugr/v1beta",
        params=params,
        wrapper_method=_to_hugr_ref,
        nexus_client=get_nexus_client(),
    )


def _to_hugr_ref(page_json: dict[str, Any]) -> DataframableList[HUGRRef]:
    """Convert JSON response dict to a list of HUGRRefs."""

    hugr_refs: DataframableList[HUGRRef] = DataframableList([])

    for hugr_data in page_json["data"]:
        project_id = hugr_data["relationships"]["project"]["data"]["id"]
        project_details = next(
            proj for proj in page_json["included"] if proj["id"] == project_id
        )
        project = ProjectRef(
            id=project_id,
            annotations=Annotations.from_dict(project_details["attributes"]),
            contents_modified=project_details["attributes"]["contents_modified"],
            archived=project_details["attributes"]["archived"],
        )

        hugr_refs.append(
            HUGRRef(
                id=UUID(hugr_data["id"]),
                annotations=Annotations.from_dict(hugr_data["attributes"]),
                project=project,
            )
        )
    return hugr_refs


def get(
    *,
    id: Union[UUID, str, None] = None,
    name_like: str | None = None,
    creator_email: list[str] | None = None,
    project: ProjectRef | None = None,
    properties: PropertiesDict | None = None,
    created_before: datetime | None = None,
    created_after: datetime | None = datetime(day=1, month=1, year=2023),
    modified_before: datetime | None = None,
    modified_after: datetime | None = None,
    sort_filters: list[SortFilterEnum] | None = None,
    page_number: int | None = None,
    page_size: int | None = None,
    scope: ScopeFilterEnum | None = None,
) -> HUGRRef:
    """
    Get a single HUGR using filters. Throws an exception if the filters do
    not match exactly one object.
    """

    if id:
        return _fetch_by_id(hugr_id=id, scope=scope)

    return get_all(
        name_like=name_like,
        creator_email=creator_email,
        properties=properties,
        project=project,
        created_before=created_before,
        created_after=created_after,
        modified_before=modified_before,
        modified_after=modified_after,
        sort_filters=sort_filters,
        page_number=page_number,
        page_size=page_size,
        scope=scope,
    ).try_unique_match()


@merge_properties_from_context
def upload(
    hugr_package: Package,
    name: str,
    project: ProjectRef | None = None,
    description: str | None = None,
    properties: PropertiesDict | None = None,
) -> HUGRRef:
    """Upload a HUGR to Nexus.

    N.B. HUGR support in Nexus is subject to change. Until full support is achieved any
    programs uploaded may not work in the future.
    """
    project = project or get_active_project(project_required=True)
    project = cast(ProjectRef, project)

    attributes = {"contents": _encode_hugr(hugr_package)}

    annotations = CreateAnnotations(
        name=name,
        description=description,
        properties=properties,
    ).model_dump(exclude_none=True)
    attributes.update(annotations)
    relationships = {"project": {"data": {"id": str(project.id), "type": "project"}}}

    req_dict = {
        "data": {
            "attributes": attributes,
            "relationships": relationships,
            "type": "hugr",
        }
    }

    res = get_nexus_client().post("/api/hugr/v1beta", json=req_dict)

    if res.status_code != 201:
        raise qnx_exc.ResourceCreateFailed(
            message=res.text, status_code=res.status_code
        )

    res_data_dict = res.json()["data"]

    return HUGRRef(
        id=UUID(res_data_dict["id"]),
        annotations=Annotations.from_dict(res_data_dict["attributes"]),
        project=project,
    )


@merge_properties_from_context
def update(
    ref: HUGRRef,
    name: str | None = None,
    description: str | None = None,
    properties: PropertiesDict | None = None,
) -> HUGRRef:
    """Update the annotations on a HUGRRef."""
    ref_annotations = ref.annotations.model_dump()
    annotations = Annotations(
        name=name,
        description=description,
        properties=properties if properties else PropertiesDict(),
    ).model_dump(exclude_none=True)
    ref_annotations.update(annotations)

    req_dict = {
        "data": {
            "attributes": annotations,
            "relationships": {},
            "type": "hugr",
        }
    }

    res = get_nexus_client().patch(f"/api/hugr/v1beta/{ref.id}", json=req_dict)

    if res.status_code != 200:
        raise qnx_exc.ResourceUpdateFailed(
            message=res.text, status_code=res.status_code
        )

    res_dict = res.json()["data"]

    return HUGRRef(
        id=UUID(res_dict["id"]),
        annotations=Annotations.from_dict(res_dict["attributes"]),
        project=ref.project,
    )


def _fetch_by_id(hugr_id: UUID | str, scope: ScopeFilterEnum | None) -> HUGRRef:
    """Utility method for fetching directly by a unique identifier."""

    params = Params(
        scope=scope,
    ).model_dump(by_alias=True, exclude_unset=True, exclude_none=True)

    res = get_nexus_client().get(f"/api/hugr/v1beta/{hugr_id}", params=params)

    handle_fetch_errors(res)

    res_dict = res.json()

    project_id = res_dict["data"]["relationships"]["project"]["data"]["id"]
    project_details = next(
        proj for proj in res_dict["included"] if proj["id"] == project_id
    )
    project = ProjectRef(
        id=project_id,
        annotations=Annotations.from_dict(project_details["attributes"]),
        contents_modified=project_details["attributes"]["contents_modified"],
        archived=project_details["attributes"]["archived"],
    )

    return HUGRRef(
        id=UUID(res_dict["data"]["id"]),
        annotations=Annotations.from_dict(res_dict["data"]["attributes"]),
        project=project,
    )


def _fetch_hugr_package(handle: HUGRRef) -> Package:
    """Utility method for fetching a HUGR Package from a HUGRRef."""
    res = get_nexus_client().get(f"/api/hugr/v1beta/{handle.id}")
    if res.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(message=res.text, status_code=res.status_code)

    contents: str = res.json()["data"]["attributes"]["contents"]
    return _decode_hugr(contents)


def _encode_hugr(hugr_package: Package) -> str:
    """Utility method for encoding a HUGR Package as base64-encoded string"""
    return base64.b64encode(
        hugr_package.to_bytes(
            config=ENVELOPE_CONFIG,
        )
    ).decode("utf-8")


def _decode_hugr(contents: str) -> Package:
    """Utility method for decoding a base64-encoded string into a HUGR Package"""
    hugr_envelope = base64.b64decode(contents)
    return Package.from_bytes(envelope=hugr_envelope)



"""Utlity functions for the client."""
# pylint: disable=protected-access
import http
import json
import os
from pathlib import Path
from typing import Any, Literal

from httpx import Response
from pydantic import BaseModel

import qnexus.exceptions as qnx_exc
from qnexus.config import get_config

TokenTypes = Literal["access_token", "refresh_token"]

token_file_from_type = {
    "access_token": "id.json",
    "refresh_token": "token.json",
}


def normalize_included(included: list[Any]) -> dict[str, dict[str, Any]]:
    """Convert a JSON API included array into a mapped dict of the form:
    {
        "user": {
            [user_id]: User
        },
        "project": {
            [project_id]: Project
        }
    }
    """
    included_map: dict[str, dict[str, Any]] = {}
    for item in included:
        included_map.setdefault(item["type"], {item["id"]: {}})
        included_map[item["type"]][item["id"]] = item
    return included_map


def remove_token(token_type: TokenTypes) -> None:
    """Delete a token file."""
    # Don't try to delete refresh token in Jupyterhub
    if is_jupyterhub_environment() and token_type == "refresh_token":
        return
    token_file_path = (
        Path.home() / get_config().token_path / token_file_from_type[token_type]
    )
    if token_file_path.exists():
        token_file_path.unlink()


class RefreshTokenData(BaseModel):
    """Stored refresh token data."""

    delete_version_after: str | None
    refresh_token: str


class RefreshToken(BaseModel):
    """Model for token storage file."""

    data: RefreshTokenData


class AccessTokenData(BaseModel):
    """Stored access token data."""

    access_token: str


class AccessToken(BaseModel):
    """Model for access token storage."""

    data: AccessTokenData


def read_token(token_type: TokenTypes) -> str:
    """Read a token from a file."""
    token_file_path = Path.home() / get_config().token_path
    with (token_file_path / token_file_from_type[token_type]).open(
        encoding="UTF-8"
    ) as file:
        file_contents = file.read().strip()
        if token_type == "access_token":
            return AccessToken(**json.loads(file_contents)).data.access_token
        return RefreshToken(**json.loads(file_contents)).data.refresh_token


def write_token(token_type: TokenTypes, token: str) -> None:
    """Write a token to a file."""

    # don't allow writing of refresh token in Jupyterhub
    if is_jupyterhub_environment() and token_type == "refresh_token":
        return

    token_file_path = Path.home() / get_config().token_path
    token_file_path.mkdir(parents=True, exist_ok=True)
    with (token_file_path / token_file_from_type[token_type]).open(
        encoding="UTF-8", mode="w"
    ) as file:
        if token_type == "access_token":
            file.write(
                AccessToken(data=AccessTokenData(access_token=token)).model_dump_json()
            )
            return
        file.write(
            RefreshToken(
                data=RefreshTokenData(refresh_token=token, delete_version_after=None)
            ).model_dump_json()
        )


def consolidate_error(res: Response, description: str) -> None:
    """Consolidate as much error-checking of response"""
    # check if token has expired or is generally unauthorized
    resp_json = res.json()
    if res.status_code == http.HTTPStatus.UNAUTHORIZED:
        raise qnx_exc.AuthenticationError(
            (
                f"Authorization failure attempting: {description}."
                f"\n\nServer Response: {resp_json}"
            )
        )
    if res.status_code != http.HTTPStatus.OK:
        raise qnx_exc.AuthenticationError(
            f"HTTP error attempting: {description}.\n\nServer Response: {resp_json}"
        )


def handle_fetch_errors(res: Response) -> None:
    """Handle errors related to a fetch request."""

    if res.status_code == 404:
        raise qnx_exc.ZeroMatches()

    if res.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(message=res.text, status_code=res.status_code)


def is_jupyterhub_environment() -> bool:
    """Check if the module is running in the Nexus JupyterHub."""
    if os.environ.get("JUPYTERHUB_USER"):
        return True
    return False



"""Client API for role-based-access-control roles and assignments in Nexus."""
from typing import Literal

from pydantic import EmailStr

import qnexus.client.teams as team_client
import qnexus.client.users as user_client
import qnexus.exceptions as qnx_exc
from qnexus.client import get_nexus_client
from qnexus.models import Role, RoleInfo
from qnexus.models.references import BaseRef, DataframableList, TeamRef

Permission = Literal["ASSIGN", "DELETE", "WRITE", "READ"]
RoleName = Literal["Administrator", "Contributor", "Reader", "Maintainer"]


def get_all() -> DataframableList[Role]:
    """Get the definitions of possible role-based access control assignments."""
    res = get_nexus_client().get(
        "/api/roles/v1beta",
    )

    if res.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(message=res.text, status_code=res.status_code)

    return DataframableList(
        [
            Role(
                id=role["id"],
                name=role["attributes"]["name"],
                description=role["attributes"]["description"],
                permissions=str(role["attributes"]["permissions"]),
            )
            for role in res.json()["data"]
        ]
    )


def get(name: RoleName) -> Role:
    """
    Get a single ``Role`` by name.
    """
    for item in get_all():
        if item.name == name:
            return item

    raise qnx_exc.NoUniqueMatch()


def assignments(resource_ref: BaseRef) -> DataframableList[RoleInfo]:
    """Check the assignments on a particular resource."""

    res = get_nexus_client().get(
        f"/api/resources/v1beta/{resource_ref.id}/assignments",
    )

    if res.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(message=res.text, status_code=res.status_code)

    roles_dict = {str(role.id): role for role in get_all()}

    res_assignments = res.json()["data"]["attributes"]

    role_infos: DataframableList[RoleInfo] = DataframableList([])

    for user_role_assignment in res_assignments["user_role_assignments"]:
        role_infos.append(
            RoleInfo(
                assignment_type="user",
                assignee=user_client._fetch_by_id(  # pylint: disable=protected-access
                    user_id=user_role_assignment["user_id"]
                ),
                role=roles_dict[user_role_assignment["role_id"]],
            )
        )
    for team_role_assignment in res_assignments["team_role_assignments"]:
        role_infos.append(
            RoleInfo(
                assignment_type="team",
                assignee=team_client._fetch_by_id(  # pylint: disable=protected-access
                    team_id=team_role_assignment["team_id"]
                ),
                role=roles_dict[team_role_assignment["role_id"]],
            )
        )
    for public_role_assignment in res_assignments["public_role_assignments"]:
        role_infos.append(
            RoleInfo(
                assignment_type="public",
                assignee=None,
                role=roles_dict[public_role_assignment["role_id"]],
            )
        )
    return role_infos


def assign_team(resource_ref: BaseRef, team: TeamRef, role: RoleName | Role) -> None:
    """Assign a role-based access control assignment to a team."""
    if isinstance(role, str):
        role = get(role)

    req_dict = {
        "data": {
            "attributes": {
                "role_id": str(role.id),
                "resource_id": str(resource_ref.id),
                "team_id": str(team.id),
            },
            "relationships": None,
            "type": "team_role_assignment",
        }
    }

    res = get_nexus_client().post(
        "/api/assignments/v1beta/team",
        json=req_dict,
    )

    if res.status_code != 201:
        raise qnx_exc.ResourceUpdateFailed(
            message=res.text, status_code=res.status_code
        )


def assign_user(
    resource_ref: BaseRef, user_email: EmailStr, role: RoleName | Role
) -> None:
    """Assign a role-based access control assignment to a user."""
    if isinstance(role, str):
        role = get(role)

    req_dict = {
        "data": {
            "attributes": {
                "role_id": str(role.id),
                "resource_id": str(resource_ref.id),
                "user": {"user_email": user_email},
            },
            "relationships": None,
            "type": "user_role_assignment",
        }
    }

    res = get_nexus_client().post(
        "/api/assignments/v1beta/user",
        json=req_dict,
    )

    if res.status_code != 201:
        raise qnx_exc.ResourceUpdateFailed(
            message=res.text, status_code=res.status_code
        )



"""CLI for qnexus."""
import click

from ..client import auth as _auth


@click.command()
def login():
    """Log in to quantinuum nexus using your web browser."""
    click.echo(_auth.login())


@click.command()
def logout():
    """Log out of quantinuum nexus."""
    click.echo(_auth.logout())



"""Client API for projects in Nexus."""
# pylint: disable=redefined-builtin
import random
from datetime import datetime
from typing import Any, Literal, Union, cast
from uuid import UUID

import pandas as pd

import qnexus.exceptions as qnx_exc
from qnexus.client import get_nexus_client
from qnexus.client.nexus_iterator import NexusIterator
from qnexus.client.utils import handle_fetch_errors
from qnexus.context import get_active_project
from qnexus.models import Property
from qnexus.models.annotations import Annotations, CreateAnnotations, PropertiesDict
from qnexus.models.filters import (  # PropertiesFilter, # Not yet implemented
    ArchivedFilter,
    CreatorFilter,
    FuzzyNameFilter,
    PaginationFilter,
    ScopeFilter,
    ScopeFilterEnum,
    SortFilter,
    SortFilterEnum,
    TimeFilter,
)
from qnexus.models.references import DataframableList, ProjectRef

# Colour-blind friendly colours from https://www.nature.com/articles/nmeth.1618
_COLOURS = ["#e69f00", "#56b4e9", "#009e73", "#f0e442", "#0072b2", "#d55e00", "#cc79a7"]


class Params(
    ScopeFilter,
    SortFilter,
    PaginationFilter,
    FuzzyNameFilter,
    CreatorFilter,
    # PropertiesFilter, # Not yet implemented
    TimeFilter,
    ArchivedFilter,
):
    """Params for filtering projects"""


def get_all(  # pylint: disable=too-many-positional-arguments
    name_like: str | None = None,
    creator_email: list[str] | None = None,
    created_before: datetime | None = None,
    created_after: datetime | None = datetime(day=1, month=1, year=2023),
    modified_before: datetime | None = None,
    modified_after: datetime | None = None,
    is_archived: bool = False,
    sort_filters: list[SortFilterEnum] | None = None,
    page_number: int | None = None,
    page_size: int | None = None,
    scope: ScopeFilterEnum | None = None,
) -> NexusIterator[ProjectRef]:
    """Get a NexusIterator over projects with optional filters."""

    params = Params(
        name_like=name_like,
        creator_email=creator_email,
        created_before=created_before,
        created_after=created_after,
        modified_before=modified_before,
        modified_after=modified_after,
        is_archived=is_archived,
        sort=SortFilter.convert_sort_filters(sort_filters),
        page_number=page_number,
        page_size=page_size,
        scope=scope,
    ).model_dump(
        by_alias=True,
        exclude_unset=True,
        exclude_none=True,
    )

    return NexusIterator(
        resource_type="Project",
        nexus_url="/api/projects/v1beta",
        params=params,
        wrapper_method=_to_projectref,
        nexus_client=get_nexus_client(),
    )


def _to_projectref(data: dict[str, Any]) -> DataframableList[ProjectRef]:
    return DataframableList(
        [
            ProjectRef(
                id=project["id"],
                annotations=Annotations.from_dict(project["attributes"]),
                contents_modified=project["attributes"]["contents_modified"],
                archived=project["attributes"]["archived"],
            )
            for project in data["data"]
        ]
    )


def get(
    *,
    id: Union[str, UUID, None] = None,
    name_like: str | None = None,
    creator_email: list[str] | None = None,
    created_before: datetime | None = None,
    created_after: datetime | None = datetime(day=1, month=1, year=2023),
    modified_before: datetime | None = None,
    modified_after: datetime | None = None,
    is_archived: bool = False,
    sort_filters: list[SortFilterEnum] | None = None,
    page_number: int | None = None,
    page_size: int | None = None,
    scope: ScopeFilterEnum | None = None,
) -> ProjectRef:
    """
    Get a single project using filters. Throws an exception if the filters do
    not match exactly one object.
    """
    if id:
        return _fetch_by_id(id, scope=scope)

    return get_all(
        name_like=name_like,
        creator_email=creator_email,
        created_before=created_before,
        created_after=created_after,
        modified_before=modified_before,
        modified_after=modified_after,
        is_archived=is_archived,
        sort_filters=sort_filters,
        page_number=page_number,
        page_size=page_size,
        scope=scope,
    ).try_unique_match()


def get_or_create(
    name: str,
    description: str | None = None,
    properties: PropertiesDict | None = None,
) -> ProjectRef:
    """Get a project reference if the projects exists (by name),
    otherwise create a new project using the supplied description and properties."""
    annotations = CreateAnnotations(
        name=name,
        description=description,
        properties=properties,
    )
    try:
        return get(name_like=annotations.name)
    except qnx_exc.ZeroMatches:
        return create(
            name=annotations.name,
            description=annotations.description,
            properties=annotations.properties,
        )


def _fetch_by_id(project_id: UUID | str, scope: ScopeFilterEnum | None) -> ProjectRef:
    """Utility method for fetching directly by a unique identifier."""
    params = Params(scope=scope).model_dump(
        by_alias=True, exclude_unset=True, exclude_none=True
    )

    res = get_nexus_client().get(f"/api/projects/v1beta/{project_id}", params=params)

    handle_fetch_errors(res)

    res_dict = res.json()

    return ProjectRef(
        id=res_dict["data"]["id"],
        annotations=Annotations.from_dict(res_dict["data"]["attributes"]),
        contents_modified=res_dict["data"]["attributes"]["contents_modified"],
        archived=res_dict["data"]["attributes"]["archived"],
    )


def create(
    name: str,
    description: str | None = None,
    properties: PropertiesDict | None = None,
) -> ProjectRef:
    """Create a new project in Nexus."""
    attributes = {}
    annotations = CreateAnnotations(
        name=name,
        description=description,
        properties=properties,
    ).model_dump(exclude_none=True)
    attributes.update(annotations)

    req_dict = {
        "data": {
            "attributes": attributes,
            "relationships": {},
            "type": "project",
        }
    }

    res = get_nexus_client().post("/api/projects/v1beta", json=req_dict)

    if res.status_code != 201:
        raise qnx_exc.ResourceCreateFailed(
            message=res.text, status_code=res.status_code
        )

    res_data_dict = res.json()["data"]

    return ProjectRef(
        id=UUID(res_data_dict["id"]),
        annotations=Annotations.from_dict(res_data_dict["attributes"]),
        contents_modified=res_data_dict["attributes"]["contents_modified"],
        archived=res_data_dict["attributes"]["archived"],
    )


def add_property(
    name: str,
    property_type: Literal["bool", "int", "float", "string"],
    project: ProjectRef | None = None,
    description: str | None = None,
    required: bool = False,
) -> None:
    """Add a property definition to a project."""
    project = project or get_active_project(project_required=True)
    assert project, "ProjectRef required."

    # For now required to add properties in a seperate API step
    props_req_dict = {
        "data": {
            "attributes": {
                "name": name,
                "description": description,
                "property_type": property_type,
                "required": required,
                "color": random.choice(_COLOURS),
            },
            "relationships": {
                "project": {"data": {"id": str(project.id), "type": "project"}}
            },
            "type": "property",
        }
    }
    props_res = get_nexus_client().post(
        "/api/property_definitions/v1beta", json=props_req_dict
    )

    if props_res.status_code != 201:
        raise qnx_exc.ResourceCreateFailed(
            message=props_res.text, status_code=props_res.status_code
        )


def get_properties(project: ProjectRef | None = None) -> DataframableList[Property]:
    """Get the property definitions for the Project."""
    # This will get all the properties (even if there are more than the page size),
    # but could be refactored if needed.

    project = project or get_active_project(project_required=True)
    assert project, "ProjectRef required."

    all_project_properties_iterator = NexusIterator(
        resource_type="Property",
        nexus_url="/api/property_definitions/v1beta",
        params={"filter[project][id]": str(project.id)},
        wrapper_method=_to_property,
        nexus_client=get_nexus_client(),
    )

    return all_project_properties_iterator.list()


def _to_property(data: dict[str, Any]) -> DataframableList[Property]:
    return DataframableList(
        [
            Property(
                annotations=Annotations.from_dict(property_dict["attributes"]),
                property_type=property_dict["attributes"]["property_type"],
                required=property_dict["attributes"]["required"],
                color=property_dict["attributes"]["color"],
                id=property_dict["id"],
            )
            for property_dict in data["data"]
        ]
    )


def summarize(project: ProjectRef | None = None) -> pd.DataFrame:
    """Summarize the current state of a project."""
    import qnexus.client.jobs as jobs_client  # pylint: disable=import-outside-toplevel

    project = project or get_active_project(project_required=True)
    project = cast(ProjectRef, project)

    all_jobs = jobs_client.get_all(project=project).list()

    return pd.DataFrame(
        {
            "project": project.annotations.name,
            "total_jobs": len(all_jobs),
            "pending_jobs": len(
                [
                    job
                    for job in all_jobs
                    if job.last_status in jobs_client.WAITING_STATUS
                ]
            ),
            "cancelled_jobs": len(
                [
                    job
                    for job in all_jobs
                    if job.last_status == jobs_client.StatusEnum.CANCELLED
                ]
            ),
            "errored_jobs": len(
                [
                    job
                    for job in all_jobs
                    if job.last_status == jobs_client.StatusEnum.ERROR
                ]
            ),
            "completed_jobs": len(
                [
                    job
                    for job in all_jobs
                    if job.last_status == jobs_client.StatusEnum.COMPLETED
                ]
            ),
        },
        index=[0],
    )


def update(
    project: ProjectRef,
    name: str | None = None,
    description: str | None = None,
    archive: bool = False,
) -> ProjectRef:
    """Update the details of a project."""
    req_dict = {
        "data": {
            "attributes": {
                "name": name,
                "description": description,
                "archived": archive,
            },
            "type": "project",
            "relationships": {},
        }
    }

    res = get_nexus_client().patch(f"/api/projects/v1beta/{project.id}", json=req_dict)

    if res.status_code != 200:
        raise qnx_exc.ResourceUpdateFailed(
            message=res.text, status_code=res.status_code
        )

    res_data_dict = res.json()["data"]

    return ProjectRef(
        id=UUID(res_data_dict["id"]),
        annotations=Annotations.from_dict(res_data_dict["attributes"]),
        contents_modified=res_data_dict["attributes"]["contents_modified"],
        archived=res_data_dict["attributes"]["archived"],
    )


def delete(project: ProjectRef) -> None:
    """Delete a project and all associated data in Nexus.
    Project must be archived first.
    WARNING: this will delete all data associated with the project.
    """
    res = get_nexus_client().delete(
        url=f"/api/projects/v1beta/{project.id}", params={"scope": "user"}
    )

    if res.status_code != 204:
        raise qnx_exc.ResourceDeleteFailed(
            message=res.text, status_code=res.status_code
        )



"""Models for use by the client."""
from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import Any, Literal, Optional
from uuid import UUID

import pandas as pd
from pydantic import BaseModel, ConfigDict, field_validator
from pytket.backends.backendinfo import BackendInfo
from quantinuum_schemas.models.backend_config import (
    AerConfig,
    AerStateConfig,
    AerUnitaryConfig,
    BackendConfig,
    BraketConfig,
    IBMQConfig,
    IBMQEmulatorConfig,
    ProjectQConfig,
    QuantinuumConfig,
    QulacsConfig,
)
from quantinuum_schemas.models.backend_info import StoredBackendInfo

from qnexus.models.annotations import Annotations
from qnexus.models.references import TeamRef, UserRef
from qnexus.models.utils import assert_never

__all__ = [
    "AerConfig",
    "AerStateConfig",
    "AerUnitaryConfig",
    "BackendConfig",
    "BraketConfig",
    "IBMQConfig",
    "IBMQEmulatorConfig",
    "ProjectQConfig",
    "QuantinuumConfig",
    "QulacsConfig",
    "StoredBackendInfo",
]


class CredentialIssuer(str, Enum):
    """An Issuer of credentials."""

    QUANTINUUM = "Quantinuum"
    QISKIT = "Qiskit"
    BRAKET = "Braket"


class Credential(BaseModel):
    """A saved credential for a backend provider/issuer."""

    name: str
    backend_issuer: CredentialIssuer
    is_default_for_issuer: bool
    submitted_time: datetime
    id: str

    def df(self) -> pd.DataFrame:
        """Convert to a pandas DataFrame."""
        return pd.DataFrame(
            {
                "name": self.name,
                "issuer": self.backend_issuer,
                "is_default_for_issuer": self.is_default_for_issuer,
                "created": self.submitted_time,
                "id": self.id,
            },
            index=[0],
        )


class Device(BaseModel):
    """A device in Nexus."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    backend_name: str
    device_name: Optional[str]
    nexus_hosted: bool
    stored_backend_info: StoredBackendInfo

    @property
    def backend_info(self) -> BackendInfo:
        """The BackendInfo for the Device."""
        return self.stored_backend_info.to_pytket_backend_info()

    def df(self) -> pd.DataFrame:
        """Present in a pandas DataFrame."""
        return pd.DataFrame(
            {
                "backend_name": self.backend_name,
                "device_name": self.device_name,
                "nexus_hosted": self.nexus_hosted,
                "backend_info": self.stored_backend_info.to_pytket_backend_info(),
            },
            index=[0],
        )

    @field_validator("backend_name")
    @classmethod
    def convert_backend_name(cls, v: str) -> str:
        """Convert the internal name for QuantinuumBackend."""
        if v == "EmulatorEnabledQuantinuumBackend":
            return "QuantinuumBackend"
        return v


class Quota(BaseModel):
    """A quota in Nexus."""

    name: str
    description: str
    usage: float
    quota: float | str

    def df(self) -> pd.DataFrame:
        """Convert to a pandas DataFrame."""
        return pd.DataFrame(self.model_dump(), index=[0])


class Role(BaseModel):
    """A role for use in RBAC assignments."""

    id: UUID
    name: str
    description: str
    permissions: str
    type: Literal["role"] = "role"

    def df(self) -> pd.DataFrame:
        """Convert to a pandas DataFrame."""
        return pd.DataFrame(
            {
                "name": self.name,
                "description": self.description,
                "permissions": self.permissions,
                "id": self.id,
            },
            index=[0],
        )


class RoleInfo(BaseModel):
    """Information on a role assigned on a resource."""

    assignment_type: Literal["user", "team", "public"]
    assignee: TeamRef | UserRef | None
    role: Role

    def df(self) -> pd.DataFrame:
        """Convert to a pandas DataFrame."""

        assignee_name: str | None = None
        match self.assignee:
            case TeamRef():
                assignee_name = self.assignee.name
            case UserRef():
                assignee_name = self.assignee.display_name
            case None:
                assignee_name = None
            case _:
                assert_never(self.assignee)

        return pd.DataFrame(
            {
                "assignment_type": self.assignment_type,
                "assignee": assignee_name,
                "role": self.role.name,
            },
            index=[0],
        )


class Property(BaseModel):
    """A property definition."""

    annotations: Annotations
    property_type: str
    required: bool
    color: str
    id: UUID

    def df(self) -> pd.DataFrame:
        """Convert to a pandas DataFrame."""
        return self.annotations.df().join(
            pd.DataFrame(
                {
                    "property_type": self.property_type,
                    "required": self.required,
                    "color": self.color,
                },
                index=[0],
            )
        )


class IssuerEnum(str, Enum):
    """Backend issuers in Nexus."""

    QUANTINUUM = "QUANTINUUM"
    IBMQ = "IBMQ"
    QISKIT = "QISKIT"
    BRAKET = "BRAKET"
    PROJECTQ = "PROJECTQ"
    QULACS = "QULACS"


def issuer_enum_to_config_str(issuer: IssuerEnum) -> list[str]:
    """Convert an IssuerEnum to a list of BackendConfig names."""

    match issuer:
        case IssuerEnum.QUANTINUUM:
            return ["QuantinuumConfig"]
        case IssuerEnum.IBMQ:
            return ["IBMQConfig"]
        case IssuerEnum.QISKIT:
            return [
                "IBMQEmulatorConfig",
                "AerConfig",
                "AerStateConfig",
                "AerUnitaryConfig",
            ]
        case IssuerEnum.PROJECTQ:
            return ["ProjectQConfig"]
        case IssuerEnum.QULACS:
            return ["QulacsConfig"]
        case IssuerEnum.BRAKET:
            return ["BraketConfig"]
        case _:
            assert_never(issuer)



""" A class for iterating over paginated API endpoints, collecting any
iterated type T into a python Iterator."""
from __future__ import annotations

from typing import Any, Callable, Dict, Generic, Iterator, List, TypeVar

import httpx
import pandas as pd

import qnexus.exceptions as qnx_exc
from qnexus.models.references import Dataframable, DataframableList

T = TypeVar("T", bound=Dataframable)


class NexusIterator(Generic[T], Iterator[T]):
    """An object that can be used to summarize or iterate through a filter query made to
    the Nexus database."""

    def __init__(  # pylint: disable=too-many-positional-arguments
        self,
        resource_type: str,
        nexus_url: str,
        params: Dict[str, Any],
        wrapper_method: Callable[[dict[str, Any]], List[T]],
        nexus_client: httpx.Client,
    ) -> None:
        self.nexus_client = nexus_client
        self.resource_type = resource_type
        self.nexus_url = nexus_url
        self.wrapper = wrapper_method
        self.current_page: int = 0
        self.params = params
        self._cached_list: DataframableList[T] | None = None
        self._current_page_subiterator: Iterator[T] = iter([])

    def __iter__(self) -> Iterator[T]:
        """Return the Iterator."""
        return self

    def __next__(self) -> T:
        """Get the next element of the Iterator."""
        try:
            return next(self._current_page_subiterator)
        except StopIteration as exc:
            self.params["page[number]"] = (self.current_page,)
            res = self.nexus_client.get(url=self.nexus_url, params=self.params)

            self._handle_errors(res)
            self.current_page += 1

            if res.json()["data"]:
                self._current_page_subiterator = iter(self.wrapper(res.json()))
                return next(self._current_page_subiterator)
            raise StopIteration from exc

    def list(self) -> DataframableList[T]:
        """Collapse into RefList."""
        if not self._cached_list:
            self._cached_list = DataframableList([])
            for item in self:
                self._cached_list.append(item)
        return self._cached_list

    def df(self) -> pd.DataFrame:
        """List and present in a pandas DataFrame."""
        return self.list().df()

    def count(self) -> int:
        """Count the items that match the filter."""

        res = self.nexus_client.get(
            url=self.nexus_url + "/meta/count", params=self.params
        )

        self._handle_errors(res)

        res_dict = res.json()
        return res_dict["count"]

    def summarize(self) -> pd.DataFrame:
        """Present in a pandas DataFrame."""
        return pd.DataFrame(
            {"resource": self.resource_type, "total_count": self.count()}, index=[0]
        )

    def _handle_errors(self, res: httpx.Response) -> None:
        """Consolidated error handling."""
        if res.status_code != 200:
            raise qnx_exc.ResourceFetchFailed(
                message=res.text, status_code=res.status_code
            )

    def try_unique_match(self) -> T:
        """Utility function for expecting a single match on the filter."""
        match_count = self.count()

        if match_count > 1:
            raise qnx_exc.NoUniqueMatch()
        if match_count == 0:
            raise qnx_exc.ZeroMatches()
        return self.list()[0]



"""Client API for execution in Nexus."""

from typing import Union, cast

from pytket.backends.backendinfo import BackendInfo
from pytket.backends.backendresult import BackendResult
from pytket.backends.status import StatusEnum
from quantinuum_schemas.models.result import QSysResult

import qnexus.exceptions as qnx_exc
from qnexus.client import circuits as circuit_api
from qnexus.client import get_nexus_client
from qnexus.client import hugr as hugr_api
from qnexus.context import get_active_project, merge_properties_from_context
from qnexus.models import BackendConfig, StoredBackendInfo
from qnexus.models.annotations import Annotations, CreateAnnotations, PropertiesDict
from qnexus.models.language import Language
from qnexus.models.references import (
    CircuitRef,
    DataframableList,
    ExecuteJobRef,
    ExecutionProgram,
    ExecutionResultRef,
    HUGRRef,
    JobType,
    ProjectRef,
    ResultType,
    WasmModuleRef,
)
from qnexus.models.utils import assert_never


@merge_properties_from_context
def start_execute_job(  # pylint: disable=too-many-arguments, too-many-locals, too-many-positional-arguments
    circuits: Union[ExecutionProgram, list[ExecutionProgram]],
    n_shots: list[int] | list[None],
    backend_config: BackendConfig,
    name: str,
    description: str = "",
    properties: PropertiesDict | None = None,
    project: ProjectRef | None = None,
    valid_check: bool = True,
    postprocess: bool = False,
    noisy_simulator: bool = True,
    language: Language = Language.AUTO,
    seed: int | None = None,
    credential_name: str | None = None,
    wasm_module: WasmModuleRef | None = None,
    user_group: str | None = None,
) -> ExecuteJobRef:
    """
    Submit an execute job to be run in Nexus. Returns an ``ExecuteJobRef``
    object which can be used to check the job's status.  See ``qnexus.execute``
    for a utility method that waits for the results and returns them.
    """
    project = project or get_active_project(project_required=True)
    project = cast(ProjectRef, project)

    program_ids = (
        [str(p.id) for p in circuits]
        if isinstance(circuits, list)
        else [str(circuits.id)]
    )

    if len(n_shots) != len(program_ids):
        raise ValueError("Number of programs must equal number of n_shots.")

    attributes_dict = CreateAnnotations(
        name=name,
        description=description,
        properties=properties,
    ).model_dump(exclude_none=True)
    attributes_dict.update(
        {
            "job_type": "execute",
            "definition": {
                "job_definition_type": "execute_job_definition",
                "backend_config": backend_config.model_dump(),
                "user_group": user_group,
                "valid_check": valid_check,
                "postprocess": postprocess,
                "noisy_simulator": noisy_simulator,
                "language": language.value
                if isinstance(language, Language)
                else language,
                "seed": seed,
                "wasm_module_id": str(wasm_module.id) if wasm_module else None,
                "credential_name": credential_name,
                "items": [
                    {"circuit_id": program_id, "n_shots": n_shot}
                    for program_id, n_shot in zip(program_ids, n_shots)
                ],
            },
        }
    )
    relationships = {
        "project": {"data": {"id": str(project.id), "type": "project"}},
        "circuits": {
            "data": [
                {"id": str(program_id), "type": "circuit"} for program_id in program_ids
            ]
        },
    }
    req_dict = {
        "data": {
            "attributes": attributes_dict,
            "relationships": relationships,
            "type": "job",
        }
    }

    resp = get_nexus_client().post(
        "/api/jobs/v1beta",
        json=req_dict,
    )
    if resp.status_code != 202:
        raise qnx_exc.ResourceCreateFailed(
            message=resp.text, status_code=resp.status_code
        )

    return ExecuteJobRef(
        id=resp.json()["data"]["id"],
        annotations=Annotations.from_dict(resp.json()["data"]["attributes"]),
        job_type=JobType.EXECUTE,
        last_status=StatusEnum.SUBMITTED,
        last_message="",
        project=project,
        backend_config_store=backend_config,
    )


def _results(
    execute_job: ExecuteJobRef,
    allow_incomplete: bool = False,
) -> DataframableList[ExecutionResultRef]:
    """Get the results from an execute job."""

    resp = get_nexus_client().get(f"/api/jobs/v1beta/{execute_job.id}")

    if resp.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(
            message=resp.text, status_code=resp.status_code
        )
    resp_data = resp.json()["data"]
    job_status = resp_data["attributes"]["status"]["status"]

    if job_status != "COMPLETED" and not allow_incomplete:
        raise qnx_exc.ResourceFetchFailed(message=f"Job status: {job_status}")

    execute_results: DataframableList[ExecutionResultRef] = DataframableList([])

    for item in resp_data["attributes"]["definition"]["items"]:
        if item["status"]["status"] == "COMPLETED":
            result_type: ResultType

            match item["result_type"]:
                case ResultType.QSYS:
                    result_type = ResultType.QSYS
                case ResultType.PYTKET:
                    result_type = ResultType.PYTKET
                case _:
                    assert_never(item["result_type"])

            result_ref = ExecutionResultRef(
                id=item["result_id"],
                annotations=execute_job.annotations,
                project=execute_job.project,
                result_type=result_type,
            )

            execute_results.append(result_ref)

    return execute_results


def _fetch_pytket_execution_result(
    result_ref: ExecutionResultRef,
) -> tuple[BackendResult, BackendInfo, CircuitRef]:
    """Get the results for an execute job item."""
    assert result_ref.result_type == ResultType.PYTKET, "Incorrect result type"

    res = get_nexus_client().get(f"/api/results/v1beta/{result_ref.id}")
    if res.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(message=res.text, status_code=res.status_code)

    res_dict = res.json()

    input_circuit_id = res_dict["data"]["relationships"]["circuit"]["data"]["id"]

    input_circuit = circuit_api._fetch_by_id(  # pylint: disable=protected-access
        input_circuit_id,
        scope=None,
    )

    results_data = res_dict["data"]["attributes"]

    results_dict = {k: v for k, v in results_data.items() if v != [] and v is not None}

    backend_result = BackendResult.from_dict(results_dict)

    backend_info_data = next(
        data for data in res_dict["included"] if data["type"] == "backend_snapshot"
    )
    backend_info = StoredBackendInfo(
        **backend_info_data["attributes"]
    ).to_pytket_backend_info()

    return (backend_result, backend_info, input_circuit)


def _fetch_qsys_execution_result(
    result_ref: ExecutionResultRef,
) -> tuple[QSysResult, BackendInfo, HUGRRef]:
    """Get the results of a next-gen Qsys execute job."""
    assert result_ref.result_type == ResultType.QSYS, "Incorrect result type"

    res = get_nexus_client().get(f"/api/qsys_results/v1beta/{result_ref.id}")

    if res.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(message=res.text, status_code=res.status_code)

    res_dict = res.json()

    input_hugr_id = res_dict["data"]["relationships"]["hugr_module"]["data"]["id"]

    input_hugr = hugr_api._fetch_by_id(  # pylint: disable=protected-access
        input_hugr_id,
        scope=None,
    )

    qsys_result = QSysResult(res_dict["data"]["attributes"]["results"])

    backend_info_data = next(
        data for data in res_dict["included"] if data["type"] == "backend_snapshot"
    )
    backend_info = StoredBackendInfo(
        **backend_info_data["attributes"]
    ).to_pytket_backend_info()

    return (
        qsys_result,
        backend_info,
        input_hugr,
    )
